@@ -7,12 +7,11 @@ import pickle as pkl
 import click
 import json
 
-
+from itertools import product
 from pexpect import spawn, EOF
-from typing import List
+from typing import List, Sequence
 
-
-DEFAULT_RECIPE = r"""Bootstrap: {bootstrap}
+RECIPE_TEMPLATE = r"""Bootstrap: {bootstrap}
 From: {bootstrap_from}
 
 %post
@@ -59,14 +58,12 @@ chmod -R 777 /code
 chmod -R 777 /results
 chmod -R 777 /anaconda3"""
 
-
 # arguments for setting up the package environment in singularity
 DEFAULT_BOOTSTRAP = "docker"
 DEFAULT_BOOTSTRAP_FROM = "nvidia/cuda:11.3.1-cudnn8-devel-ubuntu16.04"
 DEFAULT_APT_PACKAGES = ("unzip", "htop", "wget",
                         "git", "vim", "cmake", "gcc", "g++")
 
-
 # arguments that describe the conda environment to build
 DEFAULT_ANACONDA_VERSION = "2021.05"
 DEFAULT_PYTHON_VERSION = "3.8"
@@ -74,44 +71,37 @@ DEFAULT_ENV_NAME = "venv"
 DEFAULT_ENV_CREATE_ARGUMENTS = \
     "pytorch torchvision torchaudio cudatoolkit=11.3 -c pytorch"
 
-
 # a default location for the singularity image and singularity recipe
-DEFAULT_LOCAL_RECIPE = "experiment.recipe"
+DEFAULT_RECIPE = "experiment.recipe"
 DEFAULT_LOCAL_IMAGE = "experiment.sif"
-DEFAULT_REMOTE_RECIPE = "experiment.recipe"
 DEFAULT_REMOTE_IMAGE = "experiment.sif"
 
-
 # commands that are run before and after apt packages are installed
 DEFAULT_BEFORE_APT_COMMANDS = ()
 DEFAULT_POST_APT_COMMANDS = ()
 
-
 # commands that are run before and after the conda environment is created
 DEFAULT_BEFORE_ENV_COMMANDS = ()
 DEFAULT_POST_ENV_COMMANDS = ()
 
-
 # information about how to sync code before running an experiment
 DEFAULT_SYNC_WITH = ()
 DEFAULT_SYNC_TARGET = ()
 DEFAULT_EXCLUDE_FROM_SYNC = ()
 
-
-# a default command that may be run in the container
-DEFAULT_INIT_COMMANDS = ()
-
+# commands that are run at various stages of initialization
+DEFAULT_SINGULARITY_INIT_COMMANDS = ()
+DEFAULT_SINGULARITY_OPTIONS = ()
+DEFAULT_SLURM_INIT_COMMANDS = ()
 
 # a template for running experiment commands in the container
-SINGULARITY_EXEC_TEMPLATE = \
-    "singularity exec --nv -w {image} bash -c \"{singularity_command}\""
-
+SINGULARITY_EXEC_TEMPLATE = "singularity exec --nv -w \
+{singularity_options} {image} bash -c \"{singularity_command}\""
 
 # a template for launching an experiment using a slurm scheduler
 SLURM_SRUN_TEMPLATE = "sbatch --cpus-per-task={num_cpus} \
-    --gres=gpu:{num_gpus} --mem={memory}g \
-    --time={num_hours}:00:00 -p {partition} --wrap=\'{slurm_command}\'"
-
+--gres=gpu:{num_gpus} --mem={memory}g \
+--time={num_hours}:00:00 -p {partition} --wrap=\'{slurm_command}\'"
 
 # credentials for logging in to the remote host using ssh
 DEFAULT_SSH_USERNAME = "username"
@@ -120,11 +110,16 @@ DEFAULT_SSH_HOST = "matrix.ml.cmu.edu"
 DEFAULT_SSH_PORT = 22
 DEFAULT_SSH_SLEEP_SECONDS = 0.005
 
-
 # the amount of seconds to wait by default for a command ran by pexpect
 DEFAULT_PROCESS_TIMEOUT = 10800
 
 
+def add_separator(elements: Sequence[str]):
+    for i, element in enumerate(elements):
+        yield element + ("" if i == len(elements) - 1 else (
+            " " if element.strip().endswith("&") else " && "))
+
+
 class ExperimentConfig(object):
     """Create an experiment singularity image that manages packages and
     runs experiments in a reproducible and distributable container,
@@ -145,15 +140,12 @@ class ExperimentConfig(object):
         an integer representing the port on the remote machine to use
         when connecting to the machine to launch jobs.
 
-    local_recipe: str
+    recipe: str
         the location on the disk to write a singularity recipe file
         which will be used later to build a singularity image.
     local_image: str
         the location on the disk to write a singularity image, which will
         be launched when running experiments.
-    remote_recipe: str
-        the location on the host to write a singularity recipe file
-        which will be used later to build a singularity image.
     remote_image: str
         the location on the host to write a singularity image, which will
         be launched when running experiments.
@@ -204,9 +196,15 @@ class ExperimentConfig(object):
         a string representing the file pattern of files to exclude when
         synchronizing code with the singularity image.
 
-    init_commands: List[str]
+    slurm_init_commands: List[str]
         a list of strings representing commands that are run within the
-        container before starting an experiment.
+        slurm node before starting a singularity container.
+    singularity_options: List[str]
+        a list of strings that is passed to the singularity exec command, 
+        which may be used to specify mount points and gpu options.
+    singularity_init_commands: List[str]
+        a list of strings representing commands that are run within the
+        singularity container before starting an experiment.
 
     """
 
@@ -214,9 +212,8 @@ class ExperimentConfig(object):
                  ssh_password: str = DEFAULT_SSH_PASSWORD,
                  ssh_host: str = DEFAULT_SSH_HOST,
                  ssh_port: int = DEFAULT_SSH_PORT,
-                 local_recipe: str = DEFAULT_LOCAL_RECIPE,
+                 recipe: str = DEFAULT_RECIPE,
                  local_image: str = DEFAULT_LOCAL_IMAGE,
-                 remote_recipe: str = DEFAULT_REMOTE_RECIPE,
                  remote_image: str = DEFAULT_REMOTE_IMAGE,
                  before_apt_commands: List[str] = DEFAULT_BEFORE_APT_COMMANDS,
                  post_apt_commands: List[str] = DEFAULT_POST_APT_COMMANDS,
@@ -232,7 +229,10 @@ class ExperimentConfig(object):
                  sync_with: List[str] = DEFAULT_SYNC_WITH,
                  sync_target: List[str] = DEFAULT_SYNC_TARGET,
                  exclude_from_sync: List[str] = DEFAULT_EXCLUDE_FROM_SYNC,
-                 init_commands: List[str] = DEFAULT_INIT_COMMANDS):
+                 slurm_init_commands: List[str] = DEFAULT_SLURM_INIT_COMMANDS,
+                 singularity_options: List[str] = DEFAULT_SINGULARITY_OPTIONS,
+                 singularity_init_commands:
+                 List[str] = DEFAULT_SINGULARITY_INIT_COMMANDS):
         """Create an experiment singularity image that manages packages and
         runs experiments in a reproducible and distributable container,
         using a set of package configurations for each experiment.
@@ -252,15 +252,12 @@ class ExperimentConfig(object):
             an integer representing the port on the remote machine to use
             when connecting to the machine to launch jobs.
 
-        local_recipe: str
+        recipe: str
             the location on the disk to write a singularity recipe file
             which will be used later to build a singularity image.
         local_image: str
             the location on the disk to write a singularity image, which will
             be launched when running experiments.
-        remote_recipe: str
-            the location on the host to write a singularity recipe file
-            which will be used later to build a singularity image.
         remote_image: str
             the location on the host to write a singularity image, which will
             be launched when running experiments.
@@ -311,9 +308,15 @@ class ExperimentConfig(object):
             a string representing the file pattern of files to exclude when
             synchronizing code with the singularity image.
 
-        init_commands: List[str]
+        slurm_init_commands: List[str]
             a list of strings representing commands that are run within the
-            container before starting an experiment.
+            slurm node before starting a singularity container.
+        singularity_options: List[str]
+            a list of strings that is passed to the singularity exec command, 
+            which may be used to specify mount points and gpu options.
+        singularity_init_commands: List[str]
+            a list of strings representing commands that are run within the
+            singularity container before starting an experiment.
 
         """
 
@@ -324,9 +327,8 @@ class ExperimentConfig(object):
         self.ssh_port = ssh_port
 
         # locations for a singularity recipe and image to be written
-        self.local_recipe = local_recipe
+        self.recipe = recipe
         self.local_image = local_image
-        self.remote_recipe = remote_recipe
         self.remote_image = remote_image
 
         # additional commands to run when building the singularity image
@@ -351,10 +353,12 @@ class ExperimentConfig(object):
         self.sync_target = sync_target
         self.exclude_from_sync = exclude_from_sync
 
-        # arguments that specify how to sync code before an experiment
-        self.init_commands = init_commands
+        # commands that are run at various stages of initialization
+        self.slurm_init_commands = slurm_init_commands
+        self.singularity_options = singularity_options
+        self.singularity_init_commands = singularity_init_commands
 
-    def local_recipe_exists(self) -> bool:
+    def recipe_exists(self) -> bool:
         """Utility function that checks the local disk for whether a
         singularity recipe with the given name already exists on the disk
         at the desired location, and if so returns true.
@@ -367,7 +371,7 @@ class ExperimentConfig(object):
 
         """
 
-        return os.path.exists(self.local_recipe)  # exists at this location
+        return os.path.exists(self.recipe)  # exists at this location
 
     def local_image_exists(self) -> bool:
         """Utility function that checks the local disk for whether a
@@ -418,34 +422,6 @@ class ExperimentConfig(object):
         else:
             return True  # stat finished and the file is present
 
-    def remote_recipe_exists(self, client: paramiko.SSHClient = None) -> bool:
-        """Utility function that checks the remote host for whether a
-        singularity recipe with the given name already exists on the host
-        at the desired location, and if so returns true.
-
-        Returns:
-
-        recipe_exists: bool
-            a boolean that returns True if the singularity image with the
-            specified name already exists on the host.
-
-        """
-
-        if client is None:
-
-            # open an ssh connection to the remote host by logging in using
-            # the provided username and password for that machine
-            client = paramiko.SSHClient()
-            client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
-            client.connect(self.ssh_host, self.ssh_port,
-                           username=self.ssh_username,
-                           password=self.ssh_password,
-                           look_for_keys=False, allow_agent=False)
-
-        # open an sftp client and check if a file exists on the remote host
-        with client.open_sftp() as sftp:
-            return self.remote_path_exists(sftp, self.remote_recipe)
-
     def remote_image_exists(self, client: paramiko.SSHClient = None) -> bool:
         """Utility function that checks the remote host for whether a
         singularity image with the given name already exists on the host
@@ -460,7 +436,6 @@ class ExperimentConfig(object):
         """
 
         if client is None:
-
             # open an ssh connection to the remote host by logging in using
             # the provided username and password for that machine
             client = paramiko.SSHClient()
@@ -486,8 +461,8 @@ class ExperimentConfig(object):
                         .format(apt_packages=" ".join(self.apt_packages)))
 
         # write a singularity recipe file to the disk at the desired path
-        with open(self.local_recipe, "w") as recipe_file:
-            recipe_file.write(DEFAULT_RECIPE.format(
+        with open(self.recipe, "w") as recipe_file:
+            recipe_file.write(RECIPE_TEMPLATE.format(
                 bootstrap=self.bootstrap,
                 bootstrap_from=self.bootstrap_from,
                 before_apt_commands="\n".join(self.before_apt_commands),
@@ -508,12 +483,12 @@ class ExperimentConfig(object):
         """
 
         # if the recipe does not exist locally then write it first
-        if not self.local_recipe_exists():
+        if not self.recipe_exists():
             self.write_singularity_recipe()
 
         # build the singularity image using the singularity api
         from spython.main import Client
-        Client.build(recipe=self.local_recipe, image=self.local_image,
+        Client.build(recipe=self.recipe, image=self.local_image,
                      sudo=False, sandbox=True,
                      options=["--fakeroot", "--force"], **kwargs)
 
@@ -541,11 +516,13 @@ class ExperimentConfig(object):
 
         """
 
+        # handle multiple exclude patterns separated by whitespace
+        command = [y for x in exclude.split(" ") for y in ["--exclude", x]]
+        command.extend([source_path, destination_path])
+
         # spawn a child process that copies files to the host using rsync
         child = spawn("rsync", ["-ra" if recursive else "-a", "--progress",
-                                "--exclude", exclude,
-                                source_path,
-                                destination_path], encoding='utf-8')
+                                *command], encoding='utf-8')
 
         # print outputs of the process to the terminal but not passwords
         child.logfile_read = sys.stdout
@@ -601,12 +578,14 @@ class ExperimentConfig(object):
                 username=self.ssh_username,
                 host=self.ssh_host, path=destination_path)
 
+        # handle multiple exclude patterns separated by whitespace
+        command = [y for x in exclude.split(" ") for y in ["--exclude", x]]
+        command.extend(["-e", "ssh -p {}".format(self.ssh_port),
+                        source_path, destination_path])
+
         # spawn a child process that copies files to the host using rsync
         child = spawn("rsync", ["-ra" if recursive else "-a", "--progress",
-                                "--exclude", exclude,
-                                "-e", "ssh -p {}".format(self.ssh_port),
-                                source_path,
-                                destination_path], encoding='utf-8')
+                                *command], encoding='utf-8')
 
         # print outputs of the process to the terminal but not passwords
         child.logfile_read = sys.stdout
@@ -664,11 +643,13 @@ class ExperimentConfig(object):
 
         """
 
+        singularity_command = "".join(add_separator(
+            [". /anaconda3/etc/profile.d/conda.sh",
+             "conda activate {}".format(self.env_name)] +
+            list(self.singularity_init_commands) + list(commands)))
         return SINGULARITY_EXEC_TEMPLATE.format(
-            singularity_command=" && ".join(
-                [". /anaconda3/etc/profile.d/conda.sh",
-                 "conda activate {}".format(self.env_name)] +
-                list(self.init_commands) + list(commands)), image=image)
+            singularity_options=" ".join(self.singularity_options), 
+            singularity_command=singularity_command, image=image)
 
     def local_run(self, *commands: str, rebuild: bool = False):
         """Generate and run a command in the bash shell that starts a
@@ -686,9 +667,9 @@ class ExperimentConfig(object):
 
         """
 
-        if rebuild:  # if the image is being rebuilt delete the existing image
-            stdout = os.popen("rm -rf {} && rm {}"
-                              .format(self.local_image, self.local_recipe))
+        if rebuild:  # if being rebuilt delete the existing one
+            stdout = os.popen("rm {}; rm -rf {}"
+                              .format(self.recipe, self.local_image))
             for line in iter(stdout.readline, ""):
                 print(line, end="")  # prints even if not finished
 
@@ -701,7 +682,7 @@ class ExperimentConfig(object):
                 self.sync_with, self.sync_target, self.exclude_from_sync):
             self.local_rsync(os.path.join(sync_with, "."), os.path.join(
                 self.local_image, sync_target[1:]),
-                recursive=True, exclude=exclude_from_sync)
+                             recursive=True, exclude=exclude_from_sync)
 
         # start an experiment locally using a local singularity container
         stdout = os.popen(self.run_in_singularity(*commands,
@@ -752,12 +733,15 @@ class ExperimentConfig(object):
 
         """
 
+        slurm_command = "".join(add_separator(
+            list(self.slurm_init_commands) +
+            [self.run_in_singularity(*commands, image=image)]))
         return SLURM_SRUN_TEMPLATE.format(
             partition=partition, num_cpus=num_cpus, num_gpus=num_gpus,
-            num_hours=num_hours, memory=memory,
-            slurm_command=self.run_in_singularity(*commands, image=image))
+            num_hours=num_hours, memory=memory, slurm_command=slurm_command)
 
-    def remote_run(self, *commands: str, rebuild: bool = False,
+    def remote_run(self, *commands: str, exclude_nodes: str = None, rebuild: bool = False,
+                   sweep_params: List[str] = (), sweep_values: List[str] = (),
                    partition: str = "russ_reserved", num_cpus: int = 4,
                    num_gpus: int = 1, memory: int = 16, num_hours: int = 8):
         """Generate and run a command in the bash shell that starts a
@@ -773,6 +757,13 @@ class ExperimentConfig(object):
             a boolean that controls whether the singularity image should be
             rebuilt even if it already exists on the disk.
 
+        sweep_params: List[str]
+            list of strings representing names of a grid of parameters of
+            the specified command that will be searched and replaced.
+        sweep_values: List[str]
+            list of strings representing values of a grid of parameters of
+            the specified command that will be searched and replaced.
+
         num_cpus: int
             an integer representing the number of cpu cores that will be
             allocated by slurm to the generated slurm job.
@@ -800,8 +791,8 @@ class ExperimentConfig(object):
                        look_for_keys=False, allow_agent=False)
 
         if rebuild:  # if the image is being rebuilt delete the existing image
-            self.remote_shell("rm -rf {} && rm {}".format(
-                self.remote_image, self.remote_recipe), client=client)
+            self.remote_shell("rm -rf {}"
+                              .format(self.remote_image), client=client)
 
         # if the image does not exist on the remote then upload it first
         if rebuild or not self.remote_image_exists(client=client):
@@ -812,21 +803,38 @@ class ExperimentConfig(object):
                 self.sync_with, self.sync_target, self.exclude_from_sync):
             self.remote_rsync(os.path.join(sync_with, "."), os.path.join(
                 self.remote_image, sync_target[1:]),
-                recursive=True, exclude=exclude_from_sync,
-                source_is_remote=False, destination_is_remote=True)
+                              recursive=True, exclude=exclude_from_sync,
+                              source_is_remote=False, destination_is_remote=True)
 
-        # sleep in order to avoid sending too many commands to the server
-        time.sleep(DEFAULT_SSH_SLEEP_SECONDS)
+        # generate a template slurm command that will be run on the cluster
+        slurm_template = self.run_in_slurm(  # with the target node parameters
+            *commands, image=self.remote_image,
+            partition=partition, num_cpus=num_cpus,
+            num_gpus=num_gpus, num_hours=num_hours, memory=memory)
 
-        # generate a command to launch a remote experiment using slurm
-        stdout = client.exec_command(self.run_in_slurm(
-            *commands, image=self.remote_image, partition=partition,
-            num_cpus=num_cpus, num_gpus=num_gpus,
-            num_hours=num_hours, memory=memory), get_pty=True)[1]
+        # if we are excluding certain nodes on the cluster
+        if exclude_nodes is not None:  # then add an exclude flag to slurm
+            slurm_template = slurm_template.replace(
+                "sbatch", f"sbatch --exclude={exclude_nodes}")
 
-        # print the output from the terminal as the command runs
-        for line in iter(stdout.readline, ""):
-            print(line, end="")  # prints even if not yet finished
+        # iterate through every possible assignment of the parameters on an
+        # n-dimensional grid of hyper-parameters
+        for assignment in product(*map(lambda v: v.split(","), sweep_values)):
+
+            # fill in the slurm template with hyper-parameter values
+            slurm_command = slurm_template  # by replacing names with values
+            for param_name, param_value in zip(sweep_params, assignment):
+                slurm_command = slurm_command.replace(param_name, param_value)
+
+            # sleep to avoid sending too many commands to the server
+            time.sleep(DEFAULT_SSH_SLEEP_SECONDS)
+
+            # generate a command to launch a remote experiment using slurm
+            stdout = client.exec_command(slurm_command, get_pty=True)[1]
+
+            # print the output from the terminal as the command runs
+            for line in iter(stdout.readline, ""):
+                print(line, end="")  # prints even if not yet finished
 
     def remote_shell(self, *commands: str, client: paramiko.SSHClient = None,
                      watch: bool = False, interval: float = 1.0):
@@ -855,7 +863,6 @@ class ExperimentConfig(object):
                   .format(interval=interval, commands=commands))
 
         if client is None:
-
             # open an ssh connection to the remote host by logging in using
             # the provided username and password for that machine
             client = paramiko.SSHClient()
@@ -867,7 +874,7 @@ class ExperimentConfig(object):
 
         # generate a command to that runs on the remote machine over ssh
         time.sleep(DEFAULT_SSH_SLEEP_SECONDS)
-        commands = " && ".join(commands)
+        commands = "".join(add_separator(commands))
         stdout = client.exec_command(commands, get_pty=True)[1]
 
         # print intermediate outputs into the terminal as the command runs
@@ -965,9 +972,8 @@ def command_line_interface():
 @click.option('--ssh-password', type=str, default=None)
 @click.option('--ssh-host', type=str, default=None)
 @click.option('--ssh-port', type=int, default=None)
-@click.option('--local-recipe', type=str, default=None)
+@click.option('--recipe', type=str, default=None)
 @click.option('--local-image', type=str, default=None)
-@click.option('--remote-recipe', type=str, default=None)
 @click.option('--remote-image', type=str, default=None)
 @click.option('--before-apt-commands', type=str, default=None, multiple=True)
 @click.option('--no-before-apt-commands', is_flag=True)
@@ -991,15 +997,18 @@ def command_line_interface():
 @click.option('--no-sync-target', is_flag=True)
 @click.option('--exclude-from-sync', type=str, default=None, multiple=True)
 @click.option('--no-exclude-from-sync', is_flag=True)
-@click.option('--init-commands', type=str, default=None, multiple=True)
-@click.option('--no-init-commands', is_flag=True)
+@click.option('--slurm-init-commands', type=str, default=None, multiple=True)
+@click.option('--no-slurm-init-commands', is_flag=True)
+@click.option('--singularity-options', type=str, default=None, multiple=True)
+@click.option('--no-singularity-options', is_flag=True)
+@click.option('--singularity-init-commands', type=str, default=None, multiple=True)
+@click.option('--no-singularity-init-commands', is_flag=True)
 def set(ssh_username: str = DEFAULT_SSH_USERNAME,
         ssh_password: str = DEFAULT_SSH_PASSWORD,
         ssh_host: str = DEFAULT_SSH_HOST,
         ssh_port: int = DEFAULT_SSH_PORT,
-        local_recipe: str = DEFAULT_LOCAL_RECIPE,
+        recipe: str = DEFAULT_RECIPE,
         local_image: str = DEFAULT_LOCAL_IMAGE,
-        remote_recipe: str = DEFAULT_REMOTE_RECIPE,
         remote_image: str = DEFAULT_REMOTE_IMAGE,
         before_apt_commands: List[str] = DEFAULT_BEFORE_APT_COMMANDS,
         no_before_apt_commands: bool = False,
@@ -1023,8 +1032,13 @@ def set(ssh_username: str = DEFAULT_SSH_USERNAME,
         no_sync_target: bool = False,
         exclude_from_sync: List[str] = DEFAULT_EXCLUDE_FROM_SYNC,
         no_exclude_from_sync: bool = False,
-        init_commands: List[str] = DEFAULT_INIT_COMMANDS,
-        no_init_commands: bool = False):
+        slurm_init_commands: List[str] = DEFAULT_SLURM_INIT_COMMANDS,
+        no_slurm_init_commands: bool = False,
+        singularity_options: List[str] = DEFAULT_SLURM_INIT_COMMANDS,
+        no_singularity_options: bool = False,
+        singularity_init_commands:
+        List[str] = DEFAULT_SINGULARITY_INIT_COMMANDS,
+        no_singularity_init_commands: bool = False):
     """Assign new parameters to the persistent experiment config by loading
     the config from the disk and accepting new parameters for the config
     via a command line interface exposes to the shell.
@@ -1044,15 +1058,12 @@ def set(ssh_username: str = DEFAULT_SSH_USERNAME,
         an integer representing the port on the remote machine to use
         when connecting to the machine to launch jobs.
 
-    local_recipe: str
+    recipe: str
         the location on the disk to write a singularity recipe file
         which will be used later to build a singularity image.
     local_image: str
         the location on the disk to write a singularity image, which will
         be launched when running experiments.
-    remote_recipe: str
-        the location on the host to write a singularity recipe file
-        which will be used later to build a singularity image.
     remote_image: str
         the location on the host to write a singularity image, which will
         be launched when running experiments.
@@ -1103,9 +1114,15 @@ def set(ssh_username: str = DEFAULT_SSH_USERNAME,
         a string representing the file pattern of files to exclude when
         synchronizing code with the singularity image.
 
-    init_commands: List[str]
+    slurm_init_commands: List[str]
         a list of strings representing commands that are run within the
-        container before starting an experiment.
+        slurm node before starting a singularity container.
+    singularity_options: List[str]
+        a list of strings that is passed to the singularity exec command, 
+        which may be used to specify mount points and gpu options.
+    singularity_init_commands: List[str]
+        a list of strings representing commands that are run within the
+        singularity container before starting an experiment.
 
     """
 
@@ -1120,12 +1137,10 @@ def set(ssh_username: str = DEFAULT_SSH_USERNAME,
         if ssh_port is not None:
             config.ssh_port = ssh_port
 
-        if local_recipe is not None:
-            config.local_recipe = local_recipe
+        if recipe is not None:
+            config.recipe = recipe
         if local_image is not None:
             config.local_image = local_image
-        if remote_recipe is not None:
-            config.remote_recipe = remote_recipe
         if remote_image is not None:
             config.remote_image = remote_image
 
@@ -1169,9 +1184,16 @@ def set(ssh_username: str = DEFAULT_SSH_USERNAME,
             config.exclude_from_sync = \
                 () if no_exclude_from_sync else exclude_from_sync
 
-        if len(init_commands) > 0:
-            config.init_commands = \
-                () if no_init_commands else init_commands
+        if len(slurm_init_commands) > 0:
+            config.slurm_init_commands = \
+                () if no_slurm_init_commands else slurm_init_commands
+        if len(singularity_options) > 0:
+            config.singularity_options = \
+                () if no_singularity_options else singularity_options
+        if len(singularity_init_commands) > 0:
+            config.singularity_init_commands = \
+                () if no_singularity_init_commands \
+                    else singularity_init_commands
 
 
 @command_line_interface.command()
@@ -1179,9 +1201,8 @@ def set(ssh_username: str = DEFAULT_SSH_USERNAME,
 @click.option('--ssh-password', is_flag=True)
 @click.option('--ssh-host', is_flag=True)
 @click.option('--ssh-port', is_flag=True)
-@click.option('--local-recipe', is_flag=True)
+@click.option('--recipe', is_flag=True)
 @click.option('--local-image', is_flag=True)
-@click.option('--remote-recipe', is_flag=True)
 @click.option('--remote-image', is_flag=True)
 @click.option('--before-apt-commands', is_flag=True)
 @click.option('--post-apt-commands', is_flag=True)
@@ -1197,14 +1218,15 @@ def set(ssh_username: str = DEFAULT_SSH_USERNAME,
 @click.option('--sync-with', is_flag=True)
 @click.option('--sync-target', is_flag=True)
 @click.option('--exclude-from-sync', is_flag=True)
-@click.option('--init-commands', is_flag=True)
+@click.option('--slurm-init-commands', is_flag=True)
+@click.option('--singularity-options', is_flag=True)
+@click.option('--singularity-init-commands', is_flag=True)
 def get(ssh_username: bool = False,
         ssh_password: bool = False,
         ssh_host: bool = False,
         ssh_port: bool = False,
-        local_recipe: bool = False,
+        recipe: bool = False,
         local_image: bool = False,
-        remote_recipe: bool = False,
         remote_image: bool = False,
         before_apt_commands: bool = False,
         post_apt_commands: bool = False,
@@ -1220,7 +1242,9 @@ def get(ssh_username: bool = False,
         sync_with: bool = False,
         sync_target: bool = False,
         exclude_from_sync: bool = False,
-        init_commands: bool = False):
+        slurm_init_commands: bool = False,
+        singularity_options: bool = False,
+        singularity_init_commands: bool = False):
     """Read new parameters to the persistent experiment config by loading
     the config from the disk and accepting a set of indicators that
     specify which of the parameters to read to the terminal.
@@ -1240,15 +1264,12 @@ def get(ssh_username: bool = False,
         an integer representing the port on the remote machine to use
         when connecting to the machine to launch jobs.
 
-    local_recipe: str
+    recipe: str
         the location on the disk to write a singularity recipe file
         which will be used later to build a singularity image.
     local_image: str
         the location on the disk to write a singularity image, which will
         be launched when running experiments.
-    remote_recipe: str
-        the location on the host to write a singularity recipe file
-        which will be used later to build a singularity image.
     remote_image: str
         the location on the host to write a singularity image, which will
         be launched when running experiments.
@@ -1299,9 +1320,15 @@ def get(ssh_username: bool = False,
         a string representing the file pattern of files to exclude when
         synchronizing code with the singularity image.
 
-    init_commands: List[str]
+    slurm_init_commands: List[str]
         a list of strings representing commands that are run within the
-        container before starting an experiment.
+        slurm node before starting a singularity container.
+    singularity_options: List[str]
+        a list of strings that is passed to the singularity exec command, 
+        which may be used to specify mount points and gpu options.
+    singularity_init_commands: List[str]
+        a list of strings representing commands that are run within the
+        singularity container before starting an experiment.
 
     """
 
@@ -1316,12 +1343,10 @@ def get(ssh_username: bool = False,
         if ssh_port:
             print("ssh_port:", config.ssh_port)
 
-        if local_recipe:
-            print("local_recipe:", config.local_recipe)
+        if recipe:
+            print("recipe:", config.recipe)
         if local_image:
             print("local_image:", config.local_image)
-        if remote_recipe:
-            print("remote_recipe:", config.remote_recipe)
         if remote_image:
             print("remote_image:", config.remote_image)
 
@@ -1357,8 +1382,13 @@ def get(ssh_username: bool = False,
         if exclude_from_sync:
             print("exclude_from_sync:", config.exclude_from_sync)
 
-        if init_commands:
-            print("init_commands:", config.init_commands)
+        if slurm_init_commands:
+            print("slurm_init_commands:", config.slurm_init_commands)
+        if singularity_options:
+            print("singularity_options:", config.singularity_options)
+        if singularity_init_commands:
+            print("singularity_init_commands:",
+                  config.singularity_init_commands)
 
 
 @command_line_interface.command(
@@ -1393,10 +1423,14 @@ def local(rebuild: bool = False, commands: List[str] = ()):
 @click.option('--memory', type=int, default=16)
 @click.option('--num-hours', type=int, default=8)
 @click.option('--partition', type=str, default="russ_reserved")
+@click.option('--exclude-nodes', type=str, default=None)
+@click.option('--sweep-params', type=str, multiple=True, default=())
+@click.option('--sweep-values', type=str, multiple=True, default=())
 @click.option('--rebuild', is_flag=True)
 @click.argument('commands', type=str, nargs=-1)
-def remote(num_cpus: int = 4, num_gpus: int = 1, memory: int = 16,
-           num_hours: int = 8, partition: str = "russ_reserved",
+def remote(num_cpus: int = 4, num_gpus: int = 1, memory: int = 16, num_hours: int = 8,
+           partition: str = "russ_reserved", exclude_nodes: str = None,
+           sweep_params: List[str] = (), sweep_values: List[str] = (),
            rebuild: bool = False, commands: List[str] = ()):
     """Load the persistent experiment configuration file and launch an
     experiment remotely by loading the singularity image and syncing local
@@ -1420,6 +1454,13 @@ def remote(num_cpus: int = 4, num_gpus: int = 1, memory: int = 16,
         a string that represents the slurm partition of machines to use
         when scheduling a slurm job on the host machine.
 
+    sweep_params: List[str]
+        a list of strings representing names of a grid of parameters of the
+        specified command that will be searched and replaced.
+    sweep_values: List[str]
+        a list of strings representing values of a grid of parameters of the
+        specified command that will be searched and replaced.
+
     rebuild: bool
         a boolean that controls whether the singularity image should be
         rebuilt even if it already exists on the disk.
@@ -1432,8 +1473,9 @@ def remote(num_cpus: int = 4, num_gpus: int = 1, memory: int = 16,
     with PersistentExperimentConfig() as config:
         commands = [" ".join(commands)] if len(commands) > 0 else []
         config.remote_run(*commands, rebuild=rebuild, partition=partition,
-                          num_cpus=num_cpus, num_gpus=num_gpus,
-                          memory=memory, num_hours=num_hours)
+                          num_cpus=num_cpus, num_gpus=num_gpus, memory=memory,
+                          num_hours=num_hours, exclude_nodes=exclude_nodes,
+                          sweep_params=sweep_params, sweep_values=sweep_values)
 
 
 @command_line_interface.command(
@@ -1497,8 +1539,8 @@ def upload(recursive: bool = False, exclude: str = "",
     with PersistentExperimentConfig() as config:
         config.remote_rsync(source_path, os.path.join(
             config.remote_image, destination_path),
-            recursive=recursive, exclude=exclude,
-            source_is_remote=False, destination_is_remote=True)
+                            recursive=recursive, exclude=exclude,
+                            source_is_remote=False, destination_is_remote=True)
 
 
 @command_line_interface.command()
@@ -1554,30 +1596,32 @@ def dump(file: str = None):
     # export a dictionary containing the non private configuration info
     with PersistentExperimentConfig() as config:
         with open(file, "w") as f:
-            json.dump(dict(local_recipe=config.local_recipe,
+            json.dump(dict(recipe=config.recipe,
                            local_image=config.local_image,
-                           remote_recipe=config.remote_recipe,
                            remote_image=config.remote_image,
-                           
+
                            before_apt_commands=config.before_apt_commands,
                            post_apt_commands=config.post_apt_commands,
                            before_env_commands=config.before_env_commands,
                            post_env_commands=config.post_env_commands,
-                           
+
                            bootstrap=config.bootstrap,
                            bootstrap_from=config.bootstrap_from,
                            apt_packages=config.apt_packages,
-                           
+
                            anaconda_version=config.anaconda_version,
                            python_version=config.python_version,
                            env_name=config.env_name,
                            env_create_arguments=config.env_create_arguments,
-                           
+
                            sync_with=config.sync_with,
                            sync_target=config.sync_target,
                            exclude_from_sync=config.exclude_from_sync,
-                           
-                           init_commands=config.init_commands), f, indent=4)
+
+                           slurm_init_commands=config.slurm_init_commands,
+                           singularity_options=config.singularity_options,
+                           singularity_init_commands=
+                           config.singularity_init_commands), f, indent=4)
 
 
 @command_line_interface.command()
@@ -1600,31 +1644,31 @@ def load(file: str = None):
 
     # load a dictionary containing non private configuration info
     with PersistentExperimentConfig() as config:
-        
-        config.local_recipe = data["local_recipe"]
+        config.recipe = data["recipe"]
         config.local_image = data["local_image"]
-        config.remote_recipe = data["remote_recipe"]
         config.remote_image = data["remote_image"]
-        
+
         config.before_apt_commands = data["before_apt_commands"]
         config.post_apt_commands = data["post_apt_commands"]
         config.before_env_commands = data["before_env_commands"]
         config.post_env_commands = data["post_env_commands"]
-        
+
         config.bootstrap = data["bootstrap"]
         config.bootstrap_from = data["bootstrap_from"]
         config.apt_packages = data["apt_packages"]
-        
+
         config.anaconda_version = data["anaconda_version"]
         config.python_version = data["python_version"]
         config.env_name = data["env_name"]
         config.env_create_arguments = data["env_create_arguments"]
-        
+
         config.sync_with = data["sync_with"]
         config.sync_target = data["sync_target"]
         config.exclude_from_sync = data["exclude_from_sync"]
-        
-        config.init_commands = data["init_commands"]
+
+        config.slurm_init_commands = data["slurm_init_commands"]
+        config.singularity_options = data["singularity_options"]
+        config.singularity_init_commands = data["singularity_init_commands"]
 
 
 @command_line_interface.command()
